@@ -1,5 +1,6 @@
 import copy
 import os
+from pathlib import Path
 from typing import *
 from typing.io import BinaryIO, TextIO
 
@@ -11,6 +12,8 @@ from cognite.client.data_classes import (
     FileMetadataFilter,
     FileMetadataList,
     FileMetadataUpdate,
+    GeoLocation,
+    GeoLocationFilter,
     Label,
     LabelFilter,
     TimestampRange,
@@ -36,6 +39,7 @@ class FilesAPI(APIClient):
         data_set_ids: List[int] = None,
         data_set_external_ids: List[str] = None,
         labels: LabelFilter = None,
+        geo_location: GeoLocationFilter = None,
         source: str = None,
         created_time: Union[Dict[str, Any], TimestampRange] = None,
         last_updated_time: Union[Dict[str, Any], TimestampRange] = None,
@@ -65,6 +69,7 @@ class FilesAPI(APIClient):
             data_set_ids (List[int]): Return only files in the specified data sets with these ids.
             data_set_external_ids (List[str]): Return only files in the specified data sets with these external ids.
             labels (LabelFilter): Return only the files matching the specified label(s).
+            geo_location (GeoLocationFilter): Only include files matching the specified geographic relation.
             source (str): The source of this event.
             source_created_time (Union[Dict[str, Any], TimestampRange]): Filter for files where the sourceCreatedTime field has been set and is within the specified range.
             source_modified_time (Union[Dict[str, Any], TimestampRange]): Filter for files where the sourceModifiedTime field has been set and is within the specified range.
@@ -95,6 +100,7 @@ class FilesAPI(APIClient):
             root_asset_ids=root_asset_ids,
             asset_subtree_ids=asset_subtree_ids,
             labels=labels,
+            geo_location=geo_location,
             source=source,
             created_time=created_time,
             last_updated_time=last_updated_time,
@@ -148,9 +154,8 @@ class FilesAPI(APIClient):
             url_path=self._RESOURCE_PATH, json=file_metadata.dump(camel_case=True), params={"overwrite": overwrite}
         )
         returned_file_metadata = res.json()
-        upload_url = returned_file_metadata.pop("uploadUrl")
+        upload_url = returned_file_metadata["uploadUrl"]
         file_metadata = FileMetadata._load(returned_file_metadata)
-
         return (file_metadata, upload_url)
 
     def retrieve(self, id: Optional[int] = None, external_id: Optional[str] = None) -> Optional[FileMetadata]:
@@ -224,6 +229,7 @@ class FilesAPI(APIClient):
         data_set_ids: List[int] = None,
         data_set_external_ids: List[str] = None,
         labels: LabelFilter = None,
+        geo_location: GeoLocationFilter = None,
         source: str = None,
         created_time: Union[Dict[str, Any], TimestampRange] = None,
         last_updated_time: Union[Dict[str, Any], TimestampRange] = None,
@@ -250,6 +256,7 @@ class FilesAPI(APIClient):
             data_set_ids (List[int]): Return only files in the specified data sets with these ids.
             data_set_external_ids (List[str]): Return only files in the specified data sets with these external ids.
             labels (LabelFilter): Return only the files matching the specified label filter(s).
+            geo_location (GeoLocationFilter): Only include files matching the specified geographic relation.
             source (str): The source of this event.
             created_time (Union[Dict[str, int], TimestampRange]):  Range between two timestamps. Possible keys are `min` and `max`, with values given as time stamps in ms.
             last_updated_time (Union[Dict[str, int], TimestampRange]):  Range between two timestamps. Possible keys are `min` and `max`, with values given as time stamps in ms.
@@ -294,6 +301,14 @@ class FilesAPI(APIClient):
                 >>> c = CogniteClient()
                 >>> my_label_filter = LabelFilter(contains_all=["WELL LOG", "VERIFIED"])
                 >>> file_list = c.files.list(labels=my_label_filter)
+
+            Filter files based on geoLocation::
+
+                >>> from cognite.client import CogniteClient
+                >>> from cognite.client.data_classes import GeoLocationFilter, GeometryFilter
+                >>> c = CogniteClient()
+                >>> my_geo_location_filter = GeoLocationFilter(relation="intersects", shape=GeometryFilter(type="Point", coordinates=[35,10]))
+                >>> file_list = c.files.list(geo_location=my_geo_location_filter)
         """
         if root_asset_ids or root_asset_external_ids:
             root_asset_ids = self._process_ids(root_asset_ids, root_asset_external_ids, wrap_ids=True)
@@ -311,6 +326,7 @@ class FilesAPI(APIClient):
             root_asset_ids=root_asset_ids,
             asset_subtree_ids=asset_subtree_ids,
             labels=labels,
+            geo_location=geo_location,
             source=source,
             created_time=created_time,
             last_updated_time=last_updated_time,
@@ -458,6 +474,7 @@ class FilesAPI(APIClient):
         source_modified_time: int = None,
         data_set_id: int = None,
         labels: List[Label] = None,
+        geo_location: GeoLocation = None,
         security_categories: List[int] = None,
         recursive: bool = False,
         overwrite: bool = False,
@@ -475,6 +492,8 @@ class FilesAPI(APIClient):
             asset_ids (List[int]): No description.
             data_set_id (int): ID of the data set.
             labels (List[Label]): A list of the labels associated with this resource item.
+            geo_location (GeoLocation): The geographic metadata of the file.
+            security_categories (List[int]): Security categories to attach to this file.
             source_created_time (int): The timestamp for when the file was originally created in the source system.
             source_modified_time (int): The timestamp for when the file was last modified in the source system.
             recursive (bool): If path is a directory, upload all contained files recursively.
@@ -516,6 +535,13 @@ class FilesAPI(APIClient):
                 >>> c = CogniteClient()
                 >>> res = c.files.upload("/path/to/file", name="my_file", labels=[Label(external_id="WELL LOG")])
 
+            Upload a file with a geo_location::
+                >>> from cognite.client import CogniteClient
+                >>> from cognite.client.data_classes import GeoLocation, Geometry
+                >>> c = CogniteClient()
+                >>> geometry = Geometry(type="LineString", coordinates=[[30, 10], [10, 30], [40, 40]])
+                >>> res = c.files.upload("/path/to/file", geo_location=GeoLocation(type="Feature", geometry=geometry))
+
         """
         file_metadata = FileMetadata(
             name=name,
@@ -527,6 +553,7 @@ class FilesAPI(APIClient):
             asset_ids=asset_ids,
             data_set_id=data_set_id,
             labels=labels,
+            geo_location=geo_location,
             source_created_time=source_created_time,
             source_modified_time=source_modified_time,
             security_categories=security_categories,
@@ -576,6 +603,7 @@ class FilesAPI(APIClient):
         asset_ids: List[int] = None,
         data_set_id: int = None,
         labels: List[Label] = None,
+        geo_location: GeoLocation = None,
         source_created_time: int = None,
         source_modified_time: int = None,
         security_categories: List[int] = None,
@@ -596,8 +624,10 @@ class FilesAPI(APIClient):
             asset_ids (List[int]): No description.
             data_set_id (int): Id of the data set.
             labels (List[Label]): A list of the labels associated with this resource item.
+            geo_location (GeoLocation): The geographic metadata of the file.
             source_created_time (int): The timestamp for when the file was originally created in the source system.
             source_modified_time (int): The timestamp for when the file was last modified in the source system.
+            security_categories (List[int]): Security categories to attach to this file.
             overwrite (bool): If 'overwrite' is set to true, and the POST body content specifies a 'externalId' field,
                 fields for the file found for externalId can be overwritten. The default setting is false.
                 If metadata is included in the request body, all of the original metadata will be overwritten.
@@ -625,6 +655,7 @@ class FilesAPI(APIClient):
             asset_ids=asset_ids,
             data_set_id=data_set_id,
             labels=labels,
+            geo_location=geo_location,
             source_created_time=source_created_time,
             source_modified_time=source_modified_time,
             security_categories=security_categories,
@@ -634,17 +665,43 @@ class FilesAPI(APIClient):
             url_path=self._RESOURCE_PATH, json=file_metadata.dump(camel_case=True), params={"overwrite": overwrite}
         )
         returned_file_metadata = res.json()
-        upload_url = returned_file_metadata.pop("uploadUrl")
+        upload_url = returned_file_metadata["uploadUrl"]
         headers = {"X-Upload-Content-Type": file_metadata.mime_type}
-        self._http_client.request("PUT", upload_url, data=content, timeout=180, headers=headers)
+        self._http_client_with_retry.request(
+            "PUT", upload_url, data=content, timeout=self._config.file_transfer_timeout, headers=headers
+        )
         return FileMetadata._load(returned_file_metadata)
 
+    def retrieve_download_urls(
+        self, id: Union[int, List[int]] = None, external_id: Union[str, List[str]] = None
+    ) -> Dict[Union[int, str], str]:
+        """Get download links by id or external id
+
+        Args:
+            id (Union[int, List[int]]: Id or list of ids.
+            external_id (Union[str, List[str]]: External id or list of external ids.
+
+        Returns:
+            Dict[Union[str, int], str]: Dictionary containing download urls.
+        """
+        batch_size = 100
+        identifiers = self._process_ids(id, external_id, wrap_ids=True)
+        identifier_batches = [identifiers[n : n + batch_size] for n in range(0, len(identifiers), batch_size)]
+        tasks = [dict(url_path="/files/downloadlink", json={"items": id_batch}) for id_batch in identifier_batches]
+        tasks_summary = utils._concurrency.execute_tasks_concurrently(
+            self._post, tasks, max_workers=self._config.max_workers
+        )
+        tasks_summary.raise_compound_exception_if_failed_tasks()
+        results = tasks_summary.joined_results(unwrap_fn=lambda res: res.json()["items"])
+        return {result.get("id") or result["externalId"]: result["downloadUrl"] for result in results}
+
     def download(
-        self, directory: str, id: Union[int, List[int]] = None, external_id: Union[str, List[str]] = None
+        self, directory: Union[str, Path], id: Union[int, List[int]] = None, external_id: Union[str, List[str]] = None
     ) -> None:
         """`Download files by id or external id. <https://docs.cognite.com/api/v1/#operation/downloadLinks>`_
 
-        This method will stream all files to disk, never keeping more than 2MB of a given file in memory.
+        This method will stream all files to disk, never keeping more than 2MB in memory per worker.
+        The files will be stored in the provided directory using the name retrieved from the file metadata in CDF.
 
 
         Args:
@@ -663,12 +720,14 @@ class FilesAPI(APIClient):
                 >>> c = CogniteClient()
                 >>> c.files.download(directory="my_directory", id=[1,2,3], external_id=["abc", "def"])
         """
+        if isinstance(directory, str):
+            directory = Path(directory)
         all_ids = self._process_ids(ids=id, external_ids=external_id, wrap_ids=True)
         id_to_metadata = self._get_id_to_metadata_map(all_ids)
-        assert os.path.isdir(directory), "{} is not a directory".format(directory)
+        assert directory.is_dir(), "{} is not a directory".format(directory)
         self._download_files_to_directory(directory, all_ids, id_to_metadata)
 
-    def _get_id_to_metadata_map(self, all_ids):
+    def _get_id_to_metadata_map(self, all_ids: List[Dict]) -> Dict[Union[str, int], FileMetadata]:
         ids = [id["id"] for id in all_ids if "id" in id]
         external_ids = [id["externalId"] for id in all_ids if "externalId" in id]
 
@@ -681,7 +740,12 @@ class FilesAPI(APIClient):
 
         return id_to_metadata
 
-    def _download_files_to_directory(self, directory, all_ids, id_to_metadata):
+    def _download_files_to_directory(
+        self,
+        directory: Path,
+        all_ids: List[Dict[str, Union[int, str]]],
+        id_to_metadata: Dict[Union[str, int], FileMetadata],
+    ) -> None:
         tasks = [(directory, id, id_to_metadata) for id in all_ids]
         tasks_summary = utils._concurrency.execute_tasks_concurrently(
             self._process_file_download, tasks, max_workers=self._config.max_workers
@@ -696,21 +760,31 @@ class FilesAPI(APIClient):
             "downloadUrl"
         ]
 
-    def _process_file_download(self, directory, identifier, id_to_metadata):
-        download_link = self._get_download_link(identifier)
+    def _process_file_download(
+        self,
+        directory: Path,
+        identifier: Dict[str, Union[int, str]],
+        id_to_metadata: Dict[Union[str, int], FileMetadata],
+    ):
         id = utils._auxiliary.unwrap_identifer(identifier)
         file_metadata = id_to_metadata[id]
-        file_path = os.path.join(directory, file_metadata.name)
+        file_path = (directory / file_metadata.name).resolve()
+        file_is_in_download_directory = directory.resolve() in file_path.parents
+        if not file_is_in_download_directory:
+            raise RuntimeError(f"Resolved file path '{file_path}' is not inside download directory")
+        download_link = self._get_download_link(identifier)
         self._download_file_to_path(download_link, file_path)
 
-    def _download_file_to_path(self, download_link: str, path: str, chunk_size: int = 2 ** 21):
-        with self._http_client.request("GET", download_link, stream=True, timeout=180) as r:
-            with open(path, "wb") as f:
+    def _download_file_to_path(self, download_link: str, path: Path, chunk_size: int = 2 ** 21):
+        with self._http_client_with_retry.request(
+            "GET", download_link, stream=True, timeout=self._config.file_transfer_timeout
+        ) as r:
+            with path.open("wb") as f:
                 for chunk in r.iter_content(chunk_size=chunk_size):
                     if chunk:  # filter out keep-alive new chunks
                         f.write(chunk)
 
-    def download_to_path(self, path: str, id: int = None, external_id: str = None):
+    def download_to_path(self, path: Union[Path, str], id: int = None, external_id: str = None):
         """Download a file to a specific target.
 
         Args:
@@ -728,9 +802,10 @@ class FilesAPI(APIClient):
                 >>> c = CogniteClient()
                 >>> c.files.download_to_path("~/mydir/my_downloaded_file.txt", id=123)
         """
+        if isinstance(path, str):
+            path = Path(path)
         utils._auxiliary.assert_exactly_one_of_id_or_external_id(id, external_id)
-        dirname = os.path.dirname(path)
-        assert os.path.isdir(dirname), "{} is not a directory".format(dirname)
+        assert path.parent.is_dir(), "{} is not a directory".format(path.parent)
         identifier = self._process_ids(ids=id, external_ids=external_id, wrap_ids=True)[0]
         download_link = self._get_download_link(identifier)
         self._download_file_to_path(download_link, path)
@@ -756,5 +831,5 @@ class FilesAPI(APIClient):
         return self._download_file(download_link)
 
     def _download_file(self, download_link: str) -> bytes:
-        res = self._http_client.request("GET", download_link, timeout=180)
+        res = self._http_client_with_retry.request("GET", download_link, timeout=self._config.file_transfer_timeout)
         return res.content
